import types

import pytest

from potc.testing import provement
from potc.testing import transobj_assert


@pytest.mark.unittest
class TestTypes(provement()):
    def test_type_none(self):
        with transobj_assert(type(None)) as (obj, name):
            assert isinstance(obj, type)
            assert isinstance(None, obj)
            assert name == 'type_none'

    def test_types_class(self):
        with transobj_assert(types.FunctionType) as (obj, name):
            assert isinstance(obj, type)
            assert obj == types.FunctionType
            assert name == 'types_class'



from typing import NoReturn, TypeVar, Dict, Callable, List
from unittest import skipUnless

import pytest

from potc.testing import provement
from potc.testing import transobj_assert
from .testings import is_3_6, is_3_9


@pytest.mark.unittest
class TestTyping(provement()):
    def test_typing_items(self):
        with transobj_assert(NoReturn) as (obj, name):
            assert obj is NoReturn
            assert name == 'typing_items'

    def test_collection_types(self):
        with transobj_assert(List) as (obj, name):
            assert obj == List
            assert name == ('typing_items' if not is_3_6 else 'builtin_type')

        with transobj_assert(Dict) as (obj, name):
            assert obj == Dict
            assert name == ('typing_items' if not is_3_6 else 'builtin_type')

    def test_simple_collections(self):
        with transobj_assert(List[int]) as (obj, name):
            assert obj == List[int]
            assert name == 'typing_wrapper'

        with transobj_assert(Dict[int, List[int]]) as (obj, name):
            assert obj == Dict[int, List[int]]
            assert name == 'typing_wrapper'

    def test_callable(self):
        with transobj_assert(Callable[[], int]) as (obj, name):
            assert obj == Callable[[], int]
            assert name == 'typing_callable'

        with transobj_assert(Callable[..., int]) as (obj, name):
            assert obj == Callable[..., int]
            assert name == 'typing_callable'

        with transobj_assert(Callable[[int, str], List[int]]) as (obj, name):
            assert obj == Callable[[int, str], List[int]]
            assert name == 'typing_callable'

    @skipUnless(is_3_9, 'python 3.9 only')
    def test_advanced_general_alias(self):
        with transobj_assert(list[int]) as (obj, name):
            assert obj == list[int]
            assert name == 'typing_wrapper'

        with transobj_assert(dict[int, list[int]]) as (obj, name):
            assert obj == dict[int, list[int]]
            assert name == 'typing_wrapper'

    def test_typevar(self):
        K = TypeVar('K', bound=int, contravariant=True)
        V = TypeVar('V', int, str, covariant=True)

        with transobj_assert(K) as (obj, name):
            assert isinstance(obj, TypeVar)
            assert obj.__name__ == 'K'
            assert obj.__constraints__ == ()
            assert obj.__bound__ == int
            assert not obj.__covariant__
            assert obj.__contravariant__

            assert name == 'typing_typevar'

        with transobj_assert(V) as (obj, name):
            assert isinstance(obj, TypeVar)
            assert obj.__name__ == 'V'
            assert obj.__constraints__ == (int, str)
            assert obj.__bound__ is None
            assert obj.__covariant__
            assert not obj.__contravariant__

            assert name == 'typing_typevar'

        with transobj_assert(Dict[K, V]) as (obj, name):
            assert obj.__origin__ == (Dict if is_3_6 else dict)
            _k, _v = obj.__args__
            assert isinstance(_k, TypeVar)
            assert _k.__name__ == 'K'
            assert _k.__constraints__ == ()
            assert _k.__bound__ == int
            assert not _k.__covariant__
            assert _k.__contravariant__

            assert isinstance(_v, TypeVar)
            assert _v.__name__ == 'V'
            assert _v.__constraints__ == (int, str)
            assert _v.__bound__ is None
            assert _v.__covariant__
            assert not _v.__contravariant__

            assert name == 'typing_wrapper'



import re
import typing

from potc.fixture import Addons, rule
from potc.rules import builtin_type
from potc.rules.native.builtin import builtin_raw_type, builtin_object

try:
    from collections.abc import Callable as _CollectionCallable
except ImportError:
    _CollectionCallable = None
from collections import OrderedDict
from functools import lru_cache, wraps
from types import FunctionType
from typing import TypeVar, Callable

try:
    from types import GenericAlias
except ImportError:
    GenericAlias = None


@lru_cache()
def _get_origin_trans_map():
    return {
        value.__origin__: value
        for key, value in typing.__dict__.items()
        if getattr(value, '__origin__', None) and isinstance(value.__origin__, type)
    }


def _origin_trans(v):
    o = v.__origin__
    if GenericAlias is not None and isinstance(v, GenericAlias):
        return o
    else:
        return _get_origin_trans_map().get(o, o)


def _is_wrapper(func):
    @wraps(func)
    def _new_func(v, addon: Addons):
        if hasattr(v, '__origin__') and hasattr(v, '__args__'):
            return func(v, addon)
        else:
            addon.unprocessable()

    return _new_func


@rule()
@_is_wrapper
def typing_callable(v, addon: Addons):
    if (_CollectionCallable is not None and v.__origin__ == _CollectionCallable) or \
            (Callable is not None and v.__origin__ == Callable):
        _base = addon.val(_origin_trans(v))
        _args, _ret = v.__args__[:-1], v.__args__[-1]
        if _args == (Ellipsis,):
            return _base[v.__args__]
        else:
            return _base[list(_args), _ret]

    else:
        addon.unprocessable()


@rule()
@_is_wrapper
def typing_optional(v, addon: Addons):
    origin = _origin_trans(v)
    if origin is typing.Union:
        items, is_optional = [], False
        for arg in v.__args__:
            if arg is None or (isinstance(arg, type) and isinstance(None, arg)):
                is_optional = True
            else:
                items.append(arg)

        if is_optional:
            return addon.val(typing.Optional)[origin[tuple(items)]]
        else:
            addon.unprocessable()
    else:
        addon.unprocessable()


@rule()
@_is_wrapper
def typing_wrapper(v, addon: Addons):
    _base = addon.val(_origin_trans(v))
    if len(v.__args__) == 0:
        return _base
    elif len(v.__args__) == 1:
        return _base[v.__args__[0]]
    else:
        return _base[v.__args__]


_common_types = (type, FunctionType)
try:
    from typing import TypingMeta


    def _is_item(x):
        return isinstance(x, _common_types) or isinstance(type(x), TypingMeta)

except ImportError:
    from typing import _Final


    def _is_item(x):
        return isinstance(x, (*_common_types, _Final))

_ALL_NAMES = set(typing.__all__)
_STD_NAME = re.compile('^[A-Z][A-Za-z0-9]*$')


def _is_name(k):
    return k in _ALL_NAMES or _STD_NAME.fullmatch(k)


_TYPING_ITEMS = {value: key for key, value in typing.__dict__.items() if _is_name(key) and _is_item(value)}


@rule()
def typing_items(v, addon: Addons):
    if _is_item(v) and v in _TYPING_ITEMS.keys():
        _v_name = _TYPING_ITEMS[v]
        try:
            return addon.obj(v, _v_name)
        except (ImportError, TypeError):
            addon.from_(typing.__name__).import_(_v_name)
            return addon.raw(_v_name)
    else:
        addon.unprocessable()


@rule(type_=TypeVar)
def typing_typevar(v: TypeVar, addon: Addons):
    _kwargs = OrderedDict()
    if v.__bound__ is not None:
        _kwargs['bound'] = v.__bound__
    if v.__covariant__:
        _kwargs['covariant'] = v.__covariant__
    if v.__contravariant__:
        _kwargs['contravariant'] = v.__contravariant__

    return addon.obj(TypeVar)(v.__name__, *v.__constraints__, **_kwargs)


_typing_self_all = [
    (
        typing_typevar,
        typing_items,
        typing_callable,
        typing_optional,
        typing_wrapper,
    )
]

#: Overview:
#:      Typing types
#:
#: Items:
#:      - :class:`TypeVar`
#:      - Items in module ``typing``
#:      - :class:`Callable`
#:      - :class:`Wrapper`
typing_all = [
    (
        builtin_type,
        _typing_self_all,
        [builtin_raw_type, builtin_object]
    ),
]



import types

from potc.fixture import rule, Addons
from potc.rules import builtin_type


@rule(type_=type)
def type_none(v, addon: Addons):
    if v == type(None):
        return addon.val(type)(None)
    else:
        addon.unprocessable()


_TYPES_CLASSES = {
    name: getattr(types, name)
    for name in dir(types) if isinstance(getattr(types, name), type) and not name.startswith('_')
}
_TYPES_MAP = {
    obj.__name__: name for name, obj in _TYPES_CLASSES.items()
}


@rule(type_=type)
def types_class(v, addon: Addons):
    if v.__name__ in _TYPES_MAP:
        return getattr(addon.obj(types), _TYPES_MAP[v.__name__])
    else:
        addon.unprocessable()


types_all = [
    (type_none, types_class, builtin_type)
]
